@@ -27,6 +27,7 @@
 #include <config.h>
 #include <glib.h>
 
+#include "gtkhtml-private.h"
 #include "htmlclue.h"
 #include "htmlengine.h"
 #include "htmlengine-edit.h"
@@ -36,6 +37,9 @@
 
 #include "htmlcursor.h"
 
+static gboolean move_right (HTMLCursor *cursor);
+static gboolean move_left (HTMLCursor *cursor);
+
 /* #define _HTML_CURSOR_DEBUG */
 
 #ifdef _HTML_CURSOR_DEBUG
@@ -153,6 +157,8 @@ html_cursor_home (HTMLCursor *cursor,
 	g_return_if_fail (cursor != NULL);
 	g_return_if_fail (engine != NULL);
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->clue == NULL) {
 		cursor->object = NULL;
 		cursor->offset = 0;
@@ -205,31 +211,6 @@ forward (HTMLCursor *cursor)
 	return retval;
 }
 
-static gboolean
-forward_in_flow (HTMLCursor *cursor)
-{
-	gboolean retval;
-
-	retval = TRUE;
-	if (cursor->offset != html_object_get_length (cursor->object)) {
-		if (html_object_is_container (cursor->object)) {
-			HTMLObject *obj;
-
-			obj = cursor->object;
-			while ((retval = forward (cursor)) && cursor->object != obj)
-				;
-		} else
-			retval = html_object_cursor_forward (cursor->object, cursor);
-	} else {
-		if (html_object_next_not_slave (cursor->object))
-			retval = forward (cursor);
-		else
-			retval = FALSE;
-	}
-
-	return retval;
-}
-
 gboolean
 html_cursor_forward (HTMLCursor *cursor, HTMLEngine *engine)
 {
@@ -238,6 +219,8 @@ html_cursor_forward (HTMLCursor *cursor, HTMLEngine *engine)
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (engine != NULL, FALSE);
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -270,30 +253,6 @@ backward (HTMLCursor *cursor)
 	return retval;
 }
 
-static gboolean
-backward_in_flow (HTMLCursor *cursor)
-{
-	gboolean retval;
-
-	retval = TRUE;
-	if (cursor->offset  && html_object_is_container (cursor->object)) {
-		HTMLObject *obj;
-
-		obj = cursor->object;
-		while ((retval = backward (cursor)) && cursor->object != obj)
-			;
-	} else {
-		if (cursor->offset > 1 || !cursor->object->prev)
-			retval = html_object_cursor_backward (cursor->object, cursor);
-		else if (cursor->object->prev)
-			retval = backward (cursor);
-		else
-			retval = FALSE;
-	}
-
-	return retval;
-}
-
 gboolean
 html_cursor_backward (HTMLCursor *cursor,
 		      HTMLEngine *engine)
@@ -303,6 +262,8 @@ html_cursor_backward (HTMLCursor *cursor,
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (engine != NULL, FALSE);
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -321,12 +282,15 @@ html_cursor_up (HTMLCursor *cursor,
 {
 	HTMLCursor orig_cursor;
 	HTMLCursor prev_cursor;
+	HTMLDirection dir;
 	gint prev_x, prev_y;
 	gint x, y;
 	gint target_x;
 	gint orig_y;
 	gboolean new_line;
 
+	gtk_html_im_reset (engine->widget);
+
 	if (cursor->object == NULL) {
 		g_warning ("The cursor is in a NULL position: going home.");
 		html_cursor_home (cursor, engine);
@@ -336,6 +300,11 @@ html_cursor_up (HTMLCursor *cursor,
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
+	if (cursor->object->parent)
+		dir = html_object_get_direction (cursor->object->parent);
+	else
+		dir = HTML_DIRECTION_LTR;
+
 	html_cursor_copy (&orig_cursor, cursor);
 
 	html_object_get_cursor_base (cursor->object,
@@ -374,28 +343,52 @@ html_cursor_up (HTMLCursor *cursor,
 		if (y + cursor->object->descent - 1 < prev_y - prev_cursor.object->ascent) {
 			if (new_line) {
 				html_cursor_copy (cursor, &prev_cursor);
-				return FALSE;
+				return TRUE;
 			}
 
 			new_line = TRUE;
+			if (cursor->object->parent)
+				dir = html_object_get_direction (cursor->object->parent);
+			else
+				dir = HTML_DIRECTION_LTR;
 		}
 
-		if (new_line && x <= target_x) {
-			if (! cursor->have_target_x) {
-				cursor->have_target_x = TRUE;
-				cursor->target_x = target_x;
-			}
+		if (dir == HTML_DIRECTION_RTL) {
+			if (new_line && x >= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
 
-			/* Choose the character which is the nearest to the
-                           target X.  */
-			if (prev_y == y && target_x - x >= prev_x - target_x) {
-				cursor->object = prev_cursor.object;
-				cursor->offset = prev_cursor.offset;
-				cursor->position = prev_cursor.position;
-			}
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && x - target_x >= target_x - prev_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
 
-			debug_location (cursor);
-			return TRUE;
+				debug_location (cursor);
+				return TRUE;
+			}
+		} else {
+			if (new_line && x <= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
+
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && target_x - x >= prev_x - target_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
+
+				debug_location (cursor);
+				return TRUE;
+			}
 		}
 	}
 }
@@ -407,12 +400,15 @@ html_cursor_down (HTMLCursor *cursor,
 {
 	HTMLCursor orig_cursor;
 	HTMLCursor prev_cursor;
+	HTMLDirection dir;
 	gint prev_x, prev_y;
 	gint x, y;
 	gint target_x;
 	gint orig_y;
 	gboolean new_line;
 
+	gtk_html_im_reset (engine->widget);
+
 	if (cursor->object == NULL) {
 		g_warning ("The cursor is in a NULL position: going home.");
 		html_cursor_home (cursor, engine);
@@ -422,6 +418,11 @@ html_cursor_down (HTMLCursor *cursor,
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
+	if (cursor->object->parent)
+		dir = html_object_get_direction (cursor->object->parent);
+	else
+		dir = HTML_DIRECTION_LTR;
+
 	html_object_get_cursor_base (cursor->object,
 				     engine->painter, cursor->offset,
 				     &x, &y);
@@ -442,8 +443,13 @@ html_cursor_down (HTMLCursor *cursor,
 		prev_x = x;
 		prev_y = y;
 
-		if (! forward (cursor))
-			return FALSE;
+		if (dir == HTML_DIRECTION_RTL) {
+			if (! move_left (cursor))
+				return FALSE;
+		} else {
+			if (! move_right (cursor))
+				return FALSE;
+		}
 
 		html_object_get_cursor_base (cursor->object,
 					     engine->painter, cursor->offset,
@@ -457,28 +463,52 @@ html_cursor_down (HTMLCursor *cursor,
 		if (y - cursor->object->ascent > prev_y + prev_cursor.object->descent - 1) {
 			if (new_line) {
 				html_cursor_copy (cursor, &prev_cursor);
-				return FALSE;
+				return TRUE;
 			}
 
 			new_line = TRUE;
+			if (cursor->object->parent)
+				dir = html_object_get_direction (cursor->object->parent);
+			else
+				dir = HTML_DIRECTION_LTR;
 		}
 
-		if (new_line && x >= target_x) {
-			if (! cursor->have_target_x) {
-				cursor->have_target_x = TRUE;
-				cursor->target_x = target_x;
-			}
+		if (dir == HTML_DIRECTION_RTL) {
+			if (new_line && x <= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
 
-			/* Choose the character which is the nearest to the
-                           target X.  */
-			if (prev_y == y && x - target_x >= target_x - prev_x) {
-				cursor->object = prev_cursor.object;
-				cursor->offset = prev_cursor.offset;
-				cursor->position = prev_cursor.position;
-			}
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && target_x - x >= prev_x - target_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
 
-			debug_location (cursor);
-			return TRUE;
+				debug_location (cursor);
+				return TRUE;
+			}
+		} else {
+			if (new_line && x >= target_x) {
+				if (! cursor->have_target_x) {
+					cursor->have_target_x = TRUE;
+					cursor->target_x = target_x;
+				}
+
+				/* Choose the character which is the nearest to the
+				   target X.  */
+				if (prev_y == y && x - target_x >= target_x - prev_x) {
+					cursor->object = prev_cursor.object;
+					cursor->offset = prev_cursor.offset;
+					cursor->position = prev_cursor.position;
+				}
+
+				debug_location (cursor);
+				return TRUE;
+			}
 		}
 	}
 }
@@ -505,6 +535,8 @@ html_cursor_jump_to (HTMLCursor *cursor,
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (object != NULL, FALSE);
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -542,6 +574,8 @@ html_cursor_beginning_of_document (HTMLCursor *cursor,
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -557,6 +591,8 @@ html_cursor_end_of_document (HTMLCursor *cursor,
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -564,80 +600,6 @@ html_cursor_end_of_document (HTMLCursor *cursor,
 		;
 }
 
-gboolean
-html_cursor_end_of_line (HTMLCursor *cursor,
-			 HTMLEngine *engine)
-{
-	HTMLCursor prev_cursor;
-	gint x, y, prev_y;
-
-	g_return_val_if_fail (cursor != NULL, FALSE);
-	g_return_val_if_fail (engine != NULL, FALSE);
-	g_return_val_if_fail (HTML_IS_ENGINE (engine), FALSE);
-
-	cursor->have_target_x = FALSE;
-
-	if (engine->need_spell_check)
-		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
-
-	html_cursor_copy (&prev_cursor, cursor);
-	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
-				     &x, &prev_y);
-
-	while (1) {
-		if (! forward_in_flow (cursor))
-			return TRUE;
-
-		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
-					     &x, &y);
-
-		if (y - cursor->object->ascent > prev_y + prev_cursor.object->descent - 1) {
-			html_cursor_copy (cursor, &prev_cursor);
-			return TRUE;
-		}
-		prev_y = y;
-		html_cursor_copy (&prev_cursor, cursor);
-	}
-}
-
-gboolean
-html_cursor_beginning_of_line (HTMLCursor *cursor,
-			       HTMLEngine *engine)
-{
-	HTMLCursor prev_cursor;
-	gint x, y, prev_y;
-
-	g_return_val_if_fail (cursor != NULL, FALSE);
-	g_return_val_if_fail (engine != NULL, FALSE);
-	g_return_val_if_fail (HTML_IS_ENGINE (engine), FALSE);
-
-	cursor->have_target_x = FALSE;
-
-	if (engine->need_spell_check)
-		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
-
-	html_cursor_copy (&prev_cursor, cursor);
-	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
-				     &x, &prev_y);
-
-	while (1) {
-		if (! backward_in_flow (cursor))
-			return TRUE;
-
-		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
-					     &x, &y);
-
-		if (y + cursor->object->descent - 1 < prev_y - prev_cursor.object->ascent) {
-			html_cursor_copy (cursor, &prev_cursor);
-			return TRUE;
-		}
-
-		prev_y = y;
-		html_cursor_copy (&prev_cursor, cursor);
-	}
-}
-
-
 gint
 html_cursor_get_position (HTMLCursor *cursor)
 {
@@ -663,6 +625,8 @@ html_cursor_jump_to_position (HTMLCursor *cursor, HTMLEngine *engine, gint posit
 	g_return_if_fail (cursor != NULL);
 	g_return_if_fail (position >= 0);
 
+	gtk_html_im_reset (engine->widget);
+
 	if (engine->need_spell_check)
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
@@ -766,6 +730,8 @@ html_cursor_beginning_of_paragraph (HTMLCursor *cursor, HTMLEngine *engine)
 	gboolean rv = FALSE;
 	gint level, new_level;
 
+	gtk_html_im_reset (engine->widget);
+
 	level = html_object_get_parent_level (cursor->object);
 	flow  = cursor->object->parent;
 
@@ -803,6 +769,8 @@ html_cursor_end_of_paragraph (HTMLCursor *cursor, HTMLEngine *engine)
 	gboolean rv = FALSE;
 	gint level, new_level;
 
+	gtk_html_im_reset (engine->widget);
+
 	level = html_object_get_parent_level (cursor->object);
 	flow  = cursor->object->parent;
 
@@ -857,3 +825,281 @@ html_cursor_backward_n (HTMLCursor *cursor, HTMLEngine *e, guint n)
 
 	return rv;
 }
+
+HTMLObject *
+html_cursor_child_of (HTMLCursor *cursor, HTMLObject *parent)
+{
+	HTMLObject *child = cursor->object;
+
+	while (child) {
+		if (child->parent == parent)
+			return child;
+		child = child->parent;
+	}
+
+	return NULL;
+}
+
+static gboolean
+move_to_next_object (HTMLCursor *cursor)
+{
+	HTMLObject *next;
+
+	next = html_object_next_cursor (cursor->object, &cursor->offset);
+	if (next && next->parent) {
+		cursor->object = next;
+		cursor->position ++;
+
+		if (!html_object_is_container (next)) {
+			if (html_object_get_direction (next->parent) == HTML_DIRECTION_RTL) {
+				cursor->offset = html_object_get_right_edge_offset (next, 0);
+			} else {
+				cursor->offset = html_object_get_left_edge_offset (next, 0);
+			}
+			cursor->position += cursor->offset;
+		}
+
+		return TRUE;
+	} else
+		return FALSE;
+}
+
+static gboolean
+move_to_prev_object (HTMLCursor *cursor)
+{
+	HTMLObject *prev;
+
+	prev = html_object_prev_cursor (cursor->object, &cursor->offset);
+	if (prev && prev->parent) {
+		cursor->object = prev;
+		cursor->position --;
+
+		if (!html_object_is_container (prev)) {
+			if (html_object_get_direction (prev->parent) == HTML_DIRECTION_RTL) {
+				cursor->offset = html_object_get_left_edge_offset (prev, html_object_get_length (prev));
+			} else {
+				cursor->offset = html_object_get_right_edge_offset (prev, html_object_get_length (prev));
+			}
+			cursor->position -= cursor->offset - html_object_get_length (prev);
+		}
+
+		return TRUE;
+	} else
+		return FALSE;
+}
+
+static gboolean
+move_left (HTMLCursor *cursor)
+{
+	if (!html_object_cursor_left (cursor->object, cursor)) {
+		if (cursor->object->parent) {
+			if (html_object_get_direction (cursor->object->parent) == HTML_DIRECTION_RTL)
+				return move_to_next_object (cursor);
+			else
+				return move_to_prev_object (cursor);
+		}
+	}
+
+	return TRUE;
+}
+
+gboolean
+html_cursor_left (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	gboolean retval;
+
+	g_return_val_if_fail (cursor != NULL, FALSE);
+	g_return_val_if_fail (engine != NULL, FALSE);
+
+	gtk_html_im_reset (engine->widget);
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	cursor->have_target_x = FALSE;
+	retval = move_left (cursor);
+
+	debug_location (cursor);
+
+	return retval;
+}
+
+static gboolean
+left_in_flow (HTMLCursor *cursor)
+{
+	gboolean retval;
+
+	retval = TRUE;
+	if (cursor->offset != html_object_get_left_edge_offset (cursor->object, cursor->offset) && html_object_is_container (cursor->object)) {
+		HTMLObject *obj;
+
+		obj = cursor->object;
+		while ((retval = move_left (cursor)) && cursor->object != obj)
+			;
+	} else {
+		if (cursor->offset > 1 || !cursor->object->prev)
+			retval = html_object_cursor_left (cursor->object, cursor);
+		else if (cursor->object->prev)
+			retval = move_left (cursor);
+		else
+			retval = FALSE;
+	}
+
+	return retval;
+}
+
+static gboolean
+html_cursor_left_edge_of_line (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	HTMLCursor prev_cursor;
+	gint x, y, prev_y;
+
+	g_return_val_if_fail (cursor != NULL, FALSE);
+	g_return_val_if_fail (engine != NULL, FALSE);
+	g_return_val_if_fail (HTML_IS_ENGINE (engine), FALSE);
+
+	gtk_html_im_reset (engine->widget);
+
+	cursor->have_target_x = FALSE;
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	html_cursor_copy (&prev_cursor, cursor);
+	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
+				     &x, &prev_y);
+
+	while (1) {
+		if (! left_in_flow (cursor))
+			return TRUE;
+
+		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
+					     &x, &y);
+
+		if (y + cursor->object->descent - 1 < prev_y - prev_cursor.object->ascent) {
+			html_cursor_copy (cursor, &prev_cursor);
+			return TRUE;
+		}
+
+		prev_y = y;
+		html_cursor_copy (&prev_cursor, cursor);
+	}
+}
+
+static gboolean
+move_right (HTMLCursor *cursor)
+{
+	gboolean retval;
+
+	retval = TRUE;
+	if (!html_object_cursor_right (cursor->object, cursor)) {
+		if (cursor->object->parent) {
+			if (html_object_get_direction (cursor->object->parent) == HTML_DIRECTION_RTL)
+				return move_to_prev_object (cursor);
+			else
+				return move_to_next_object (cursor);
+		}
+	}
+	return retval;
+}
+
+gboolean
+html_cursor_right (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	gboolean retval;
+
+	g_return_val_if_fail (cursor != NULL, FALSE);
+	g_return_val_if_fail (engine != NULL, FALSE);
+
+	gtk_html_im_reset (engine->widget);
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	cursor->have_target_x = FALSE;
+	retval = move_right (cursor);
+
+	debug_location (cursor);
+
+	return retval;
+}
+
+static gboolean
+right_in_flow (HTMLCursor *cursor)
+{
+	gboolean retval;
+
+	retval = TRUE;
+	if (cursor->offset != html_object_get_right_edge_offset (cursor->object, cursor->offset)) {
+		if (html_object_is_container (cursor->object)) {
+			HTMLObject *obj;
+
+			obj = cursor->object;
+			while ((retval = move_right (cursor)) && cursor->object != obj)
+				;
+		} else
+			retval = html_object_cursor_right (cursor->object, cursor);
+	} else {
+		if (html_object_next_not_slave (cursor->object))
+			retval = move_right (cursor);
+		else
+			retval = FALSE;
+	}
+
+	return retval;
+}
+
+static gboolean
+html_cursor_right_edge_of_line (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	HTMLCursor prev_cursor;
+	gint x, y, prev_y;
+
+	g_return_val_if_fail (cursor != NULL, FALSE);
+	g_return_val_if_fail (engine != NULL, FALSE);
+	g_return_val_if_fail (HTML_IS_ENGINE (engine), FALSE);
+
+	gtk_html_im_reset (engine->widget);
+
+	cursor->have_target_x = FALSE;
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	html_cursor_copy (&prev_cursor, cursor);
+	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
+				     &x, &prev_y);
+
+	while (1) {
+		if (! right_in_flow (cursor))
+			return TRUE;
+
+		html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
+					     &x, &y);
+
+		if (y - cursor->object->ascent > prev_y + prev_cursor.object->descent - 1) {
+			html_cursor_copy (cursor, &prev_cursor);
+			return TRUE;
+		}
+		prev_y = y;
+		html_cursor_copy (&prev_cursor, cursor);
+	}
+}
+
+gboolean
+html_cursor_beginning_of_line (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	if (html_object_get_direction (cursor->object) == HTML_DIRECTION_RTL)
+		return html_cursor_right_edge_of_line (cursor, engine);
+	else
+		return html_cursor_left_edge_of_line (cursor, engine);
+}
+
+gboolean
+html_cursor_end_of_line (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	if (html_object_get_direction (cursor->object) == HTML_DIRECTION_RTL)
+		return html_cursor_left_edge_of_line (cursor, engine);
+	else
+		return html_cursor_right_edge_of_line (cursor, engine);
+}
