@@ -600,6 +600,128 @@ html_object_real_get_direction (HTMLObject *o)
 	return HTML_DIRECTION_DERIVED;
 }
 
+static gboolean
+html_object_real_cursor_forward (HTMLObject *self, HTMLCursor *cursor)
+{
+	gint len;
+
+	g_assert (self);
+	g_assert (cursor->object == self);
+
+	if (html_object_is_container (self))
+		return FALSE;
+
+	len = html_object_get_length (self);
+	if (cursor->offset < len) {
+		cursor->offset ++;
+		cursor->position ++;
+		return TRUE;
+	} else
+		return FALSE;
+}
+
+static gboolean
+html_object_real_cursor_backward (HTMLObject *self, HTMLCursor *cursor)
+{
+	HTMLObject *prev;
+
+	g_assert (self);
+	g_assert (cursor->object == self);
+
+	if (html_object_is_container (self))
+		return FALSE;
+
+	if (cursor->offset > 1 || (cursor->offset > 0 && (! (prev = html_object_prev_not_slave (self))
+							  || HTML_IS_CLUEALIGNED (prev) || !html_object_accepts_cursor (prev)))) {
+		cursor->offset --;
+		cursor->position --;
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean
+html_object_real_cursor_right (HTMLObject *self, HTMLCursor *cursor)
+{
+	HTMLDirection dir = html_object_get_direction (self);
+
+	g_assert (self);
+	g_assert (cursor->object == self);
+
+	if (html_object_is_container (self))
+		return FALSE;
+
+	if (dir == HTML_DIRECTION_LTR) {
+		gint len;
+
+		len = html_object_get_length (self);
+
+		if (cursor->offset < len) {
+			cursor->offset ++;
+			cursor->position ++;
+			return TRUE;
+		}
+	} else {
+		HTMLObject *prev;
+
+		if (cursor->offset > 1 || (cursor->offset > 0 && (! (prev = html_object_prev_not_slave (self))
+								  || HTML_IS_CLUEALIGNED (prev) || !html_object_accepts_cursor (prev)))) {
+			cursor->offset --;
+			cursor->position --;
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+static gboolean
+html_object_real_cursor_left (HTMLObject *self, HTMLCursor *cursor)
+{
+	HTMLDirection dir = html_object_get_direction (self);
+
+	g_assert (self);
+	g_assert (cursor->object == self);
+
+	if (html_object_is_container (self))
+		return FALSE;
+
+	if (dir == HTML_DIRECTION_LTR) {
+		HTMLObject *prev;
+		if (cursor->offset > 1 || (cursor->offset > 0 && (! (prev = html_object_prev_not_slave (self))
+								  || HTML_IS_CLUEALIGNED (prev) || !html_object_accepts_cursor (prev)))) {
+			cursor->offset --;
+			cursor->position --;
+			return TRUE;
+		}
+	} else {
+		gint len;
+
+		len = html_object_get_length (self);
+
+		if (cursor->offset < len) {
+			cursor->offset ++;
+			cursor->position ++;
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+static int
+html_object_real_get_right_edge_offset (HTMLObject *o, int offset)
+{
+	return html_object_get_length (o);
+}
+
+static int
+html_object_real_get_left_edge_offset (HTMLObject *o, int offset)
+{
+	return 0;
+}
+
 /* Class initialization.  */
 
 void
@@ -671,6 +793,12 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->get_engine = get_engine;
 	klass->get_clear = get_clear;
 	klass->get_direction = html_object_real_get_direction;
+	klass->cursor_forward = html_object_real_cursor_forward;
+	klass->cursor_backward = html_object_real_cursor_backward;
+	klass->cursor_left = html_object_real_cursor_left;
+	klass->cursor_right = html_object_real_cursor_right;
+	klass->get_right_edge_offset = html_object_real_get_right_edge_offset;
+	klass->get_left_edge_offset = html_object_real_get_left_edge_offset;
 }
 
 void
@@ -824,6 +952,26 @@ html_object_calc_abs_position (HTMLObject *o,
 	}
 }
 
+void
+html_object_calc_abs_position_in_frame (HTMLObject *o, int *x_return, int *y_return)
+{
+	HTMLObject *p;
+
+	g_return_if_fail (o != NULL);
+
+	*x_return = o->x;
+	*y_return = o->y;
+	
+	frame_offset (o, x_return, y_return);
+
+	for (p = o->parent; p != NULL && !html_object_is_frame (p); p = p->parent) {
+		*x_return += p->x;
+		*y_return += p->y - p->ascent;
+		
+		frame_offset (p, x_return, y_return);
+	}
+}
+
 GdkRectangle *
 html_object_get_bounds (HTMLObject *o, GdkRectangle *bounds)
 {
@@ -1384,42 +1532,25 @@ html_object_tail_not_slave (HTMLObject *self)
 gboolean
 html_object_cursor_forward (HTMLObject *self, HTMLCursor *cursor)
 {
-	gint len;
-
-	g_assert (self);
-	g_assert (cursor->object == self);
-
-	if (html_object_is_container (self))
-		return FALSE;
-
-	len = html_object_get_length (self);
-	if (cursor->offset < len) {
-		cursor->offset ++;
-		cursor->position ++;
-		return TRUE;
-	} else
-		return FALSE;
+	return (* HO_CLASS (self)->cursor_forward) (self, cursor);
 }
 
 gboolean
 html_object_cursor_backward (HTMLObject *self, HTMLCursor *cursor)
 {
-	HTMLObject *prev;
+	return (* HO_CLASS (self)->cursor_backward) (self, cursor);
+}
 
-	g_assert (self);
-	g_assert (cursor->object == self);
+gboolean
+html_object_cursor_right (HTMLObject *self, HTMLCursor *cursor)
+{
+	return (* HO_CLASS (self)->cursor_right) (self, cursor);
+}
 
-	if (html_object_is_container (self))
-		return FALSE;
-
-	if (cursor->offset > 1 || (cursor->offset > 0 && (! (prev = html_object_prev_not_slave (self))
-							  || HTML_IS_CLUEALIGNED (prev) || !html_object_accepts_cursor (prev)))) {
-		cursor->offset --;
-		cursor->position --;
-		return TRUE;
-	}
-
-	return FALSE;
+gboolean
+html_object_cursor_left (HTMLObject *self, HTMLCursor *cursor)
+{
+	return (* HO_CLASS (self)->cursor_left) (self, cursor);
 }
 
 /*********************
@@ -1935,8 +2066,16 @@ html_object_get_insert_level (HTMLObject *o)
 {
 	switch (HTML_OBJECT_TYPE (o)) {
 	case HTML_TYPE_TABLECELL:
-	case HTML_TYPE_CLUEV:
-		return 3;
+	case HTML_TYPE_CLUEV: {
+		int level = 3;
+
+		while (o && HTML_IS_CLUEV (o) && HTML_CLUE (o)->head && HTML_IS_CLUEV (HTML_CLUE (o)->head)) {
+			level ++;
+			o = HTML_CLUE (o)->head;
+		}
+
+		return level;
+	}
 	case HTML_TYPE_CLUEFLOW:
 		return 2;
 	default:
@@ -2099,3 +2238,15 @@ html_object_get_flow (HTMLObject *o)
 
 	return HTML_CLUEFLOW (o);
 }
+
+int
+html_object_get_right_edge_offset (HTMLObject *o, int offset)
+{
+	return (* HO_CLASS (o)->get_right_edge_offset) (o, offset);
+}
+
+int
+html_object_get_left_edge_offset (HTMLObject *o, int offset)
+{
+	return (* HO_CLASS (o)->get_left_edge_offset) (o, offset);
+}
